//---------------------------------------------------------------------------
// File  :  FFP_framework1.c
// Date  :  2009-10-02   (YYYY-MM-DD)
// Author:  Wolfgang Buescher  (DL4YHF)
// Location: <DL4YHF>C:\cbproj\SpecLab\fft_filter_plugins\*.*
//
// Description:
//   Some *OPTIONAL* framework for FFT-filter-plugins with a simple GUI .
//   Creates a small control window without a message loop (non-blocking !).
//   The plugin may later draw something into that window using plain old
//   windows GDI functions; see fft_filter_demo_1.c for an example .
//
// Compile with Dev-Cpp or Borland C++ Builder V4;
//         MS-C may work but never tested.
//  If [wx]Dev-C++ complains about a gazillion of "undefined references",
//         try adding the following library somewhere under 
//         'Project'..'Project Options'..'Parameters'(!!!!)...'Linker' :
//            ../../../Dev-Cpp/lib/libgdi32.a
//

#include <windows.h>
#include <math.h>

#include "fft_filter_dll.h" /* header for all FFT-filter plugin DLLs    */
#include "FFP_framework.h"  /* some framework for the FFT-filter plugin */
#include "icon_identifiers.h" /* IDI_SPECLAB_MAIN_ICON, etc etc   */

//---------------------------------------------------------------------
// Compilation switches .
//  MAY be defined under compiler options, or in a header like SWITCHES.H
//---------------------------------------------------------------------

#ifndef  SWI_PLUGIN_WINDOW_WIDTH
 #define SWI_PLUGIN_WINDOW_WIDTH 200
#endif

#ifndef  SWI_PLUGIN_WINDOW_HEIGHT
 #define SWI_PLUGIN_WINDOW_HEIGHT 130
#endif


//---------------------------------------------------------------------
// Internal prototypes and variables ..
//---------------------------------------------------------------------

  // "Calibration" parameters, may be set to the actual values
  // used in Spectrum Lab, i.e. copied from T_FftFilterInfo->dblAdcMaxInputVoltage
  // to use these values in subroutines which cannot access that structure .
double FFP_dblAdcMaxInputVoltage = 1.0; // max ADC input voltage [Volts] at the clipping point
double FFP_dblAdcInputImpedance = 50.0; // input impedance [Ohms] for some exotic conversions


/*  Declare Windows procedure  */
LRESULT CALLBACK FFP_WindowProc(HWND, UINT, WPARAM, LPARAM);

// Because the 'Window procedure' callback doesn't contain a pointer
// to the FFT-filter-plugin info, we need this ugly variable :
#define FFPFW_MAX_INSTANCES 4
static T_FftFilterInfo *FFPFW_pFilterInfo[FFPFW_MAX_INSTANCES] = {NULL,NULL,NULL,NULL};
static BOOL FFPFW_fWinClassRegistered = FALSE;
int    FFPFW_iDllInstanceCount = 0; // instance counter of the DLL

BOOL     g_fPaused = FALSE;

/*  Make the class name into a global variable  */
// ex: char szClassName[ ] = "FftFilterPlugin";
//     a bad idea if DIFFERENT plugins which use this framework
//     shall run side-by-side .
//     So, the window class name is now a member of T_FftFilterInfo;
//     pInfo->sz40WinClassName should be set by the plugin in FFP_Init() .





/***************************************************************************/
BOOL FFP_OpenControlPanel(
       T_FftFilterInfo *pInfo, // plugin info
       char *pszWindowTitle)   // text which appears in the window title
  // Use this function as a template for your own control window !
{
  WNDCLASSEX wincl;        /* Data structure for the windowclass */
  int i;

  if( pInfo->dwhPluginWindow == 0 )
   { // Create the window only if it hasn't been created yet !

     // If it's the FIRST window opened by this plugin,
     //  then register the window class,
     //    and if it fails don't try to create a window.
     if( ! FFPFW_fWinClassRegistered )
      {
        // Fill the "window class" structure, similar to a windows application
        memset( &wincl, 0, sizeof(wincl) );
        wincl.cbSize = sizeof (WNDCLASSEX);
        wincl.hInstance = (HINSTANCE)pInfo->dwhAppInstance;
        wincl.lpszClassName = pInfo->sz40WinClassName;
        wincl.lpfnWndProc = FFP_WindowProc;  /* This function is called by windows */
        wincl.style = CS_DBLCLKS;            /* Catch double-clicks */

        // About hIconSm:
        //  >  Handle to a small icon that is associated with the window class.
        //  > If this member is NULL, the system searches the icon resource
        //  > specified by the hIcon member for an icon of the appropriate
        //  > size to use as the small icon.
        // Here, we cannot access the application's main icon. Instead,
        // try to load the icon from the resource embedded in the plugin DLL.
        // This is tricky... often the plugin's window title bar had an EMPTY
        // icon. For example, none of the following attempts worked :
        // ; wincl.hIcon = LoadIcon (NULL, IDI_APPLICATION);
        // ; wincl.hIcon = LoadIcon (NULL, MAKEINTRESOURCE(IDI_SPECLAB_MAIN_ICON) );
        // ; wincl.hIconSm = LoadIcon (NULL, MAKEINTRESOURCE(IDI_SPECLAB_MAIN_ICON) );
        // ; wincl.hIcon = LoadIcon( (HINSTANCE)pInfo->dwhAppInstance, IDI_APPLICATION);
        // ; wincl.hIcon = LoadIcon( GetModuleHandle( NULL ), IDI_APPLICATION );
        // Finally gave up ... "and now for something completely different" :
        // Try to load the icon from the PLUGIN DLL which is a bit tricky .
        // Caution: pInfo->dwhPluginDll was returned by LoadLibrary, when
        //  the HOST loaded the plugin, so it's a HMODULE.
        //  But LoadImage expects a HINSTANCE, so we use a strange cast here:
        wincl.hIcon = LoadImage(
          (HINSTANCE)pInfo->dwhPluginDll, // HINSTANCE hinst, same as a HMODULE ?
           // "Handle to the module of either a DLL or executable" .
           // note the mircosoft slang: this is NOT the handle of a DLL !
           MAKEINTRESOURCE(IDI_SPECLAB_MAIN_ICON)/*LPCTSTR lpszName*/ ,
           IMAGE_ICON/*uType*/ ,
           32/*cxDesired*/,  32/*cyDesired*/,  LR_DEFAULTCOLOR/*"fuLoad"*/ );
        wincl.hCursor = LoadCursor (NULL, IDC_ARROW);
        wincl.lpszMenuName = NULL;  /* No menu */
        wincl.cbClsExtra = 0;       /* No extra bytes after the window class */
        wincl.cbWndExtra = 0;       /* structure or the window instance */
        wincl.hbrBackground = (HBRUSH)GetStockObject( WHITE_BRUSH ); // ex: COLOR_BACKGROUND

        if (!RegisterClassEx(&wincl) )
         { // If the window class has been registered before,
           // RegisterClassEx will fail.
           // It is important AND MUST NOT BE IGNORED,
           // because the 'window procedure' (callback address)
           // is part of the window class (not the window instance !!).
           // If RegisterClassEx fails, the previous plugin -which used
           // the same class name- has not unregistered the class
           // when the DLL was unloaded from memory. THIS IS A DANGEROUS
           // CONDITION, so we bail out here, and do NOT create a window :
           strcpy( pInfo->sz255ErrorText,"Error: Couldn't register window class !");
           return FALSE;
         }
        FFPFW_fWinClassRegistered = TRUE;
      } // end if( ! FFPFW_fWinClassRegistered )

     // Because the 'Window procedure' callback doesn't contain a pointer
     // to the FFT-filter-plugin info, save the pointer to the info struct
     // so we can access it in the event handlers (see FFP_WindowProc) .
     // This is relatively safe, because Spectrum Lab will always pass
     // a pointer to the SAME T_FftFilterInfo to the plugin (the address
     // won't change between two calls), so for the moment, we can live
     // with this kludge. Anyway, it should be avoided somehow.. any idea ?
     for(i=0; i<FFPFW_MAX_INSTANCES; ++i)
      { if(FFPFW_pFilterInfo[i] == NULL)
         { FFPFW_pFilterInfo[i] = pInfo;  // just a POINTER; not a STRUCTURE !
           break;
         }
      }

     // Now create main window, using size+position from the info struct
     if(  pInfo->iPluginWindowWidth<20 || pInfo->iPluginWindowWidth>1024
      || pInfo->iPluginWindowHeight<20 || pInfo->iPluginWindowHeight>789 )
      { // if window size and position looks invalid, restore defaults:
        pInfo->iPluginWindowWidth = SWI_PLUGIN_WINDOW_WIDTH;
        pInfo->iPluginWindowHeight= SWI_PLUGIN_WINDOW_HEIGHT;
      }
     if( pInfo->iPluginWindowX<-(pInfo->iPluginWindowWidth-10)
      || pInfo->iPluginWindowY<-(pInfo->iPluginWindowHeight-10) )
      { pInfo->iPluginWindowX = 10;
        pInfo->iPluginWindowY = 10;
      }
     pInfo->dwhPluginWindow = (DWORD)CreateWindow(
       pInfo->sz40WinClassName,  // lpClassName,  pointer to registered class name
       pszWindowTitle, // "lpWindowName", pointer to window name, effectively the TITLE
       // DWORD dwStyle,        window style. Tried a lot ...
       // WS_POPUPWINDOW | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX | WS_OVERLAPPED,  // normal window
       // WS_POPUP | WS_DLGFRAME ,   // window with no title, quite unusual as main window
       WS_POPUPWINDOW | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX | WS_SIZEBOX, // sizeable
       // WS_CAPTION,   // with title, but no icon and no system menu nor close symbol
       pInfo->iPluginWindowX,     // int x
       pInfo->iPluginWindowY,     // int y
       pInfo->iPluginWindowWidth, // int nWidth , window width
       pInfo->iPluginWindowHeight, // int nHeight, window height
       (HWND)pInfo->dwhMainWindow, // HWND hWndParent, handle to parent window:
                // if hWndParent is set, the window is minimized together with its parent
       NULL,    // HMENU hMenu,     handle to menu or child-window identifier
       (HINSTANCE)pInfo->dwhAppInstance, // HANDLE hInstance, handle to application instance
       NULL);   // LPVOID lpParam,  pointer to window-creation data

     /* Make the window visible on the screen */
     ShowWindow( (HWND)pInfo->dwhPluginWindow, SW_SHOW );

     return TRUE;

     // Note: there is NO MESSAGE LOOP in the plugin, because
     //       an application shall have only ONE such thing !
     //      (Having an message loop in the plugin would block
     //       FFP_Init() until the window was closed )
#if(0)
     /* Run the message loop. It will run until GetMessage() returns 0 */
     while (GetMessage (&messages, NULL, 0, 0))
      {
        /* Translate virtual-key messages into character messages */
        TranslateMessage(&messages);
        /* Send message to WindowProcedure */
        DispatchMessage(&messages);
      }
     /* The program return-value is 0 - The value that PostQuitMessage() gave */
     return messages.wParam;
#endif // no MESSAGE LOOP, instead we just return the handle to the new window:
   }
  else // window already created, just make it VISIBLE again:
   {
     ShowWindow( (HWND)pInfo->dwhPluginWindow, SW_SHOW );
   }
  return TRUE;
} // end FFP_OpenControlPanel()

/***************************************************************************/
void FFP_CloseControlPanel(T_FftFilterInfo *pInfo)
  // Closes the plugin's own control panel (window) without destroying it.
{
  int i;

  if( pInfo->dwhPluginWindow != 0 )
   { CloseWindow( (HWND)pInfo->dwhPluginWindow );
     // No-No: DestroyWindow( (HWND)pInfo->dwhPluginWindow );
     // Note: we do NOT destroy the window here,
     // but windows usually sends a WM_DESTROY message anyway. Why ?
   }

  for(i=0; i<FFPFW_MAX_INSTANCES; ++i)
   { if(FFPFW_pFilterInfo[i] == pInfo)
      { FFPFW_pFilterInfo[i] = NULL; // don't need this info any longer
      }
   }


} // end FFP_CloseControlPanel()

/***************************************************************************/
void FFP_UnregisterWndClassIfLastDllInstance(T_FftFilterInfo *pInfo)
{
  if( (pInfo!=NULL) && FFPFW_fWinClassRegistered && (FFPFW_iDllInstanceCount<=1) )
   { FFPFW_fWinClassRegistered = FALSE;  // avoid calling twice
     UnregisterClass( pInfo->sz40WinClassName,
           (HINSTANCE)pInfo->dwhAppInstance );
     // Note: if there are more than one instances of this DLL "loaded",
     // it's illegal to unregister the window class, because such a window
     // MAY STILL BE OPEN !
   }
} // end FFP_UnregisterWndClassIfLastDllInstance()

/***************************************************************************/
void FFP_DestroyControlPanel(T_FftFilterInfo *pInfo)
  // Destroys the plugin's own control panel (window) .
  // Should be called from FFP_Exit() .
  // No problem when called even though no own window was created .
{
 int i;

  if( pInfo->dwhPluginWindow != 0 )
   { DestroyWindow( (HWND)pInfo->dwhPluginWindow );
     pInfo->dwhPluginWindow = 0;   // window handle invalid now, forget it !
   }
  for(i=0; i<FFPFW_MAX_INSTANCES; ++i)
   { if(FFPFW_pFilterInfo[i] == pInfo)
      { FFPFW_pFilterInfo[i] = NULL; // don't need this info any longer
      }
   }

  // If this is the last instance of the DLL, unregister the window class:
  FFP_UnregisterWndClassIfLastDllInstance(pInfo);

} // end FFP_DestroyControlPanel()


/***************************************************************************/
LRESULT CALLBACK FFP_WindowProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 /* This function is called by the Windows function DispatchMessage() .
  *      Prototype dictated by windows API !
  *
  * 'Hidden' function input (in an ugly global variable):
  *    T_FftFilterInfo *FFPFW_pFilterInfo[?] = pointer to the filter plugin info.
  *    Many event handlers (which will be called from here) need that info,
  *    for example to exchange information with the worker thread.
  *    Any idea/suggestions how to get rid of this  ?
  */
{
  T_FftFilterInfo *pInfo = NULL;
  int i;

  for(i=0; i<FFPFW_MAX_INSTANCES; ++i) // get the FFT-filter-info for this window:
   { if(FFPFW_pFilterInfo[i] != NULL)
      { if( (DWORD)hwnd == FFPFW_pFilterInfo[i]->dwhPluginWindow )
         { pInfo = FFPFW_pFilterInfo[i];
           break;
         }
      }
   }
  if( pInfo==NULL )
   { return DefWindowProc (hwnd, message, wParam, lParam);
   }

  switch(message)  /* handle some of the messages here ... */
   {
     case WM_WINDOWPOSCHANGED:  // user change size and/or position of the window
        // Get the new window position+size, so SpecLab can save them later:
        pInfo->iPluginWindowX = ((WINDOWPOS*)lParam)->x;
        pInfo->iPluginWindowY = ((WINDOWPOS*)lParam)->y;
        pInfo->iPluginWindowWidth = ((WINDOWPOS*)lParam)->cx;
        pInfo->iPluginWindowHeight= ((WINDOWPOS*)lParam)->cy;
        return DefWindowProc(hwnd, message, wParam, lParam);

     case WM_DESTROY: // the plugin's window is being destroyed .. sniff ;)
        // About WM_DESTROY:
        // > sent to the window procedure of the window being destroyed
        // > *after* the window is removed from the screen.
        // No-No-No: PostQuitMessage(0);  /* send a WM_QUIT to the message queue */
        // It's not clear why WM_DESTROY is sent when the user clicks the
        // 'close' symbol. So, in fact the window seems to be DESTROYED when
        // closed. We must cope with with situation here by setting the
        // window handle invalid (the handle of a DESTROYED window is invalid!).
        pInfo->dwhPluginWindow = 0; // window handle invalid now, forget it !
        // Note: don't try to call UnregisterClass here, because
        // we cannot unregister a window in its own window procedure .
        break;  // -> return 0 to indicate we have processed the message .

     case WM_KEYDOWN:
        return FFP_OnKeyDown( pInfo, wParam/*nVirtKey*/, lParam/*lKeyData*/ );
        // "An application should return zero if it processes this message"

  /* case WM_COPYDATA:
        return OnWmCopydata( hWnd,  // handle of THIS window
                     (HWND)wParam,  // handle of SENDING window
         (PCOPYDATASTRUCT)lParam);  // pointer to a COPYDATASTRUCT (read-only!)
   */

     default:                      /* for messages that we don't deal with */
        return DefWindowProc (hwnd, message, wParam, lParam);
   }

  return 0;
} // end FFP_WindowProc()


//---------------------------------------------------------------------------
//  General helper functions (Spectrum Lab - related)
//---------------------------------------------------------------------------


/***************************************************************************/
double FFP_ConvertAnyAmplitudeToNormalizedPower(
          double dblAnyAmplitude,    // input value (to be converted)
             int iFromScaleUnit )    // unit of input value, SCALE_UNIT_xxxxx
   // Converts "any possible amplitude/magnitude value" INTO a power
   //          (or, precisely "something proportinal to an energy") .
   //      Main purpose: Averaging !  The inverse function, to convert the averaged value
   //          back into that "something", is UTL_ConvertNormalizedPowerToAnyAmplitude() .
   //      Also used to convert between other units, back and forth, for the display.
   // Caution, this function MAY use some ugly global variables, for example
   //          FFP_dblAdcMaxInputVoltage   and   FFP_dblAdcInputImpedance !
   //    (that's the reason why we prefer NORMALIZED powers internally)
{
 switch( iFromScaleUnit )
  {
   case SCALE_UNIT_NONE     :      // NO unit, or UNKNOWN unit :
   case SCALE_UNIT_UNKNOWN  :      // convert "something unknown" to VOLTAGE (how?)
        return dblAnyAmplitude * dblAnyAmplitude;  // assume it's a voltage; square it

   case SCALE_UNIT_NORM_V:         // voltage, just square it for some power-proportional thing
        return dblAnyAmplitude * dblAnyAmplitude;  // voltage squared for the result
   case SCALE_UNIT_PERCENT:        // voltage, but expressed as "percentage" (100 % = ADC clipping)
        dblAnyAmplitude *= 0.01;   // from "percent" to +/- 1.0 (normalized)
        return dblAnyAmplitude * dblAnyAmplitude;  // voltage squared for the result
   case SCALE_UNIT_VOLT:           // voltage (V), 'calibrated' using UTL_dblAdcMaxInputVoltage
        if( FFP_dblAdcMaxInputVoltage > 0.0 )
            dblAnyAmplitude /= FFP_dblAdcMaxInputVoltage;
        return dblAnyAmplitude * dblAnyAmplitude;  // voltage squared for the result

   case SCALE_UNIT_WATT     : // from POWER/Watts INTO something power-proportinal...
   case SCALE_UNIT_NORM_PWR : // power, normalized to 0...1 for full A/D converter input range
        // P = U*I = U/R; U=sqrt( P * R )
        return dblAnyAmplitude;   // just ignore the scaling factor in this case !

   case SCALE_UNIT_dB    :        // from dB , various reference level, all treated the same here:
   case SCALE_UNIT_dBfs  :
        return pow(10.0, dblAnyAmplitude/*here: dB*/ / 10.0 );  // dBfs -> normalized power
   case SCALE_UNIT_dBV   :
   case SCALE_UNIT_dBm   :
        return pow(10.0, dblAnyAmplitude/*here: dB*/ / 10.0 );  // dB -> normalized power

   case SCALE_UNIT_dBuV  :  // dBV ("dBuV") into normalized power (1.0 for full ADC input)
        dblAnyAmplitude = 1e-6 * pow(10.0, dblAnyAmplitude/*here: dB*/ / 20.0 );  // dBuV -> voltage in VOLTS
        if( FFP_dblAdcMaxInputVoltage > 0.0 )
            dblAnyAmplitude /= FFP_dblAdcMaxInputVoltage;
        return dblAnyAmplitude * dblAnyAmplitude;  // voltage squared for the result

   default:
        return dblAnyAmplitude * dblAnyAmplitude;
  } // end switch( iFromScaleUnit )

} // end FFP_ConvertAnyAmplitudeToNormalizedPower()

/***************************************************************************/
double FFP_ConvertNormalizedPowerToAnyAmplitude(
          double dblPowerOrEnergy, // input value (to be converted)
             int iToScaleUnit )    // unit of output value, SCALE_UNIT_xxxxx
   // This is the inverse function to FFP_ConvertAnyAmplitudeToNormalizedPower() .
{
  double dblResult;

 if( dblPowerOrEnergy < 1e-100 ) // this is illegal (for a power or energy) ...
     dblPowerOrEnergy = 1e-100;  // could have been a numeric rounding error (?)
 if( dblPowerOrEnergy > 1e100 )
     dblPowerOrEnergy = 1e100;
 // (no matter if it was an error, after this we can safely calculate the square root)

 switch( iToScaleUnit )
  {
   case SCALE_UNIT_NONE     :      // NO unit, or UNKNOWN unit :
   case SCALE_UNIT_UNKNOWN  :      // convert to "something unknown" (how?)
        return sqrt( dblPowerOrEnergy );  // back from "power" to "voltage"

   case SCALE_UNIT_NORM_V:         // voltage, just square it for some power-proportional thing
        return sqrt( dblPowerOrEnergy );  // back from "power" to "voltage"
   case SCALE_UNIT_PERCENT:        // voltage, but expressed as "percentage" (100 % = ADC clipping)
        return sqrt( dblPowerOrEnergy ) * 100.0;  // back from "power" to "voltage in percent"
   case SCALE_UNIT_VOLT:           // voltage (V), 'calibrated' using UTL_dblAdcMaxInputVoltage
        dblResult = sqrt( dblPowerOrEnergy );
        if( FFP_dblAdcMaxInputVoltage > 0.0 )
            dblResult *= FFP_dblAdcMaxInputVoltage;
        return dblResult;

   case SCALE_UNIT_WATT     :     // POWER/Watts ...
        return dblPowerOrEnergy;  // already ignored the scaling factor in this case !
   case SCALE_UNIT_NORM_PWR :     // power, normalized to 0...1 for full A/D converter input range
        return dblPowerOrEnergy;

   case SCALE_UNIT_dB    :        // from power into dB , various reference level, all treated the same here:
   case SCALE_UNIT_dBfs  :
   case SCALE_UNIT_dBV   :
   case SCALE_UNIT_dBm   :
        // Often CRASHED(!) here with a floating-point error (same old story, over and over..)
        // Trying to "run until return" with the debugger is wishful thinking; it never works.
        return 10.0 * log10(dblPowerOrEnergy);  // power(energy) -> dB, various ref levels

   case SCALE_UNIT_dBuV  :  // convert from normalized power (1.0 = max ADC input) to dBV ("dBuV")
        if( FFP_dblAdcMaxInputVoltage > 0.0 )
          dblResult *= sqrt(FFP_dblAdcMaxInputVoltage);
        dblResult = 10.0 * log10(dblPowerOrEnergy);  // power(energy) -> dBfs (full scale)
        // Now convert from dBV to dBuV :
        dblResult += 60.0;  // 10*log(1e6) = 60
        return dblResult;

   default:
        return sqrt( dblPowerOrEnergy );  // back from "power" to "voltage" or whatever linear thing it was
  } // end switch( iFromScaleUnit )

} // end FFP_ConvertNormalizedPowerToAnyAmplitude()



/***************************************************************************/
double FFP_ConvertAmplitude( double dblAmpl, int iFromUnit, int iToUnit )
  // Converts one amplitude unit to another .
  // Amplitude units are defined as SCALE_UNIT_xxx in FFT_FILTER_DLL,
  //           originally in "Utility1" (when compiling spectrum lab) .
  // Return value :  amplitude in the unit 'iToUnit' .
  // Example :
  //  FFP_ConvertAmplitude( 0.5, SCALE_UNIT_NORM_Vm // from: normalized voltage
  //                             SCALE_UNIT_dBfs)   // to : dB full scale
  //       will return something close to -6  ( "dB over full scale") .
  // Note:
  //   Unlike the utterly complex implementation in Spectrum Lab,
  //   this simple function does not support ANY conversion
  //   because it cannot access some 'calibration parameters'
  //   which are used inside Spectrum Lab to measure absolute voltages, etc.
  //   Try to avoid exotic amlitudes. Suggestion: Use SCALE_UNIT_dBfs only .
{
  double dblNormalizedPower;

  if( iFromUnit != iToUnit )  // do we really have to convert ?
   {
     dblNormalizedPower = FFP_ConvertAnyAmplitudeToNormalizedPower(dblAmpl, iFromUnit);
     dblAmpl = FFP_ConvertNormalizedPowerToAnyAmplitude( dblNormalizedPower, iToUnit );
   }
  return dblAmpl;
} // end FFP_ConvertAmplitude()

//---------------------------------------------------------------------------
//  Helper functions for FFP_OnDrawSpectrum()
//---------------------------------------------------------------------------

/***************************************************************************/
BOOL FFP_DrawSpectrum_CheckCompatibility(
        T_FFP_SPECTRUM_DRAWING_INFO *pDrawingInfo) // [in] info about SL's main spectrum display
  // Checks if the plugin is still compatible with the host (i.e. Spectrum Lab).
  // Returns TRUE if we can safely access the members in T_FFP_SPECTRUM_DRAWING_INFO,
  //  otherwise FALSE (in that case, don't access pDrawingInfo ) .
{
  if( pDrawingInfo==NULL )
   { return FALSE;   // this is not a valid pointer at all -> bail out
   }
  if( pDrawingInfo->iSizeOfStruct < sizeof(T_FFP_SPECTRUM_DRAWING_INFO) )
   { // The size of the T_FFP_SPECTRUM_DRAWING_INFO structure in the host (spectrum lab)
     // is SMALLER than the size of T_FFP_SPECTRUM_DRAWING_INFO
     // defined in "our" FFT_FILTER_DLL.H  ? ! ?
     // In this case, refuse to access anything else inside the structure,
     // to avoid access violations etc.
     // (Another case,
     //   ( pDrawingInfo->iSizeOfStruct < sizeof(T_FFP_SPECTRUM_DRAWING_INFO) )
     //  is no problem because new members are always appended to the
     //  end of the T_FFP_SPECTRUM_DRAWING_INFO structure. That way,
     //  NEW versions of Spectrum Lab will still work with OLDER plugins .
     return FALSE;   // Spectrum Lab too old, some new members in T_FFP_SPECTRUM_DRAWING_INFO may be missing there
   }
  else // should be ok to access all "known" members in pDrawingInfo ..
   { return TRUE;
   }

} // end FFP_DrawSpectrum_CheckCompatibility()


/***************************************************************************/
BOOL FFP_DrawSpectrum_FreqAndAmplitudeToScreenPos(
        T_FFP_SPECTRUM_DRAWING_INFO *pDrawingInfo, // [in] info about SL's main spectrum display
        double dblFrequency,   // [in] baseband frequency in Hertz
        double dblAmpl,        // [in] amplitude, magnitude, voltage or power..
        int    iAmplScaleUnit, // [in] scaling unit for 'amplitude, magnitude or power',
                               //      something like SCALE_UNIT_dBfs, see fft_filter_dll.h
        int *piX, int *piY )   // [out] pixel coordinate for painting and printing
   // Transforms a "physical" value pair (like a level in dB and a frequency in Hertz)
   // into a pixel coordinate for painting, plotting, or drawing
   // in the callback function 'FFP_OnDrawSpectrum' as defined in FFT_FILTER_DLL.H .
   // Return value:
   //   TRUE = 'the point is inside the visible curve area'
   //   FALSE= 'the point is outside the visible curve area'
   //          or 'something wrong with the function arguments' .
   //          In that case, x and y may be invalid, so DON'T USE THE COORDINATE.
{
  int x,y;
  double dA,dF;
  BOOL fResult = TRUE;

  if( ! FFP_DrawSpectrum_CheckCompatibility(pDrawingInfo) )
   { return FALSE;  // oops.. version conflict !
   }

  if( iAmplScaleUnit != pDrawingInfo->iAmplScaleUnit )
   { // The plugin's amplitude scale unit is not equal to the 'display' unit
     // -> convert it .
     dblAmpl = FFP_ConvertAmplitude( dblAmpl,
         iAmplScaleUnit/*iFromUnit*/, pDrawingInfo->iAmplScaleUnit/*iToUnit*/);
   }
  dA = pDrawingInfo->dblGraphAmplMax - pDrawingInfo->dblGraphAmplMin;
  if( dA > 0 )
   { dA = (dblAmpl - pDrawingInfo->dblGraphAmplMin) / dA;
   }
  else
   { dA = 0;
     fResult = FALSE;
   }
  dF = pDrawingInfo->dblBasebandFreqMax - pDrawingInfo->dblBasebandFreqMin;
  if( dF > 0 )
   { dF = (dblFrequency - pDrawingInfo->dblBasebandFreqMin) / dF;
   }
  else
   { dF = 0;
     fResult = FALSE;
   }
  // Note: LOGARITHMIC FREQUENCY SCALES ARE NOT SUPPORTED HERE (yet ?) ! !

  if( pDrawingInfo->iOrientation==0) // 0=normal=horizontal frequency scale
   { x = pDrawingInfo->iCurveAreaX1 + (int)(dF * (double)(pDrawingInfo->iCurveAreaX2-pDrawingInfo->iCurveAreaX1) + 0.5 );
     y = pDrawingInfo->iCurveAreaY2 - (int)(dA * (double)(pDrawingInfo->iCurveAreaY2-pDrawingInfo->iCurveAreaY1) + 0.5 );
   }
  else //  pDrawingInfo->iOrientation=1=rotated=vertical frequency scale
   { x = pDrawingInfo->iCurveAreaX1 + (int)(dA * (pDrawingInfo->iCurveAreaX2-pDrawingInfo->iCurveAreaX1));
     y = pDrawingInfo->iCurveAreaY2 - (int)(dF * (pDrawingInfo->iCurveAreaY2-pDrawingInfo->iCurveAreaY1));
   }
  if( piX != NULL )  // just an "in-range" check or a world -> viewport transformation ?
   { *piX = x;
   }
  if( piY != NULL )
   { *piY = y;
   }
  fResult &= (x>=pDrawingInfo->iCurveAreaX1)
          && (x<=pDrawingInfo->iCurveAreaX2)
          && (y>=pDrawingInfo->iCurveAreaY1)
          && (y<=pDrawingInfo->iCurveAreaY2);
  return fResult;

} // end FFP_DrawSpectrum_FreqAndAmplitudeToScreenPos()

/* EOF <FFP_framework1.c > */




